import asyncio

import pytest
from casty import ActorSystem, Behavior, Behaviors

from skyward.actors.messages import (
    Bootstrapped,
    Bootstrapping,
    Execute,
    InstanceBecameReady,
    InstanceDied,
    Log,
    Metric,
    Preempted,
    Running,
)


@pytest.fixture
def system():
    s = ActorSystem("test-instance")
    yield s
    asyncio.get_event_loop().run_until_complete(s.shutdown())


def collector_behavior(collected: list) -> Behavior:
    async def receive(ctx, msg):
        collected.append(msg)
        return Behaviors.same()
    return Behaviors.receive(receive)


@pytest.mark.asyncio
async def test_instance_waiting_receives_running(system):
    from skyward.actors.instance import instance_actor

    parent_msgs: list = []
    parent_ref = system.spawn(collector_behavior(parent_msgs), "parent")

    ref = system.spawn(
        instance_actor(
            instance_id="i-123",
            provider_ref=system.spawn(collector_behavior([]), "provider"),
            worker_ref=None,
            parent=parent_ref,
            _skip_tunnel=True,
        ),
        "instance",
    )

    ref.tell(Running(ip="10.0.0.1"))
    await asyncio.sleep(0.1)

    # Instance should transition to starting/bootstrapping
    # No parent notification yet (not ready)
    assert len(parent_msgs) == 0


@pytest.mark.asyncio
async def test_instance_preempted_in_waiting(system):
    from skyward.actors.instance import instance_actor

    parent_msgs: list = []
    parent_ref = system.spawn(collector_behavior(parent_msgs), "parent")

    ref = system.spawn(
        instance_actor(
            instance_id="i-123",
            provider_ref=system.spawn(collector_behavior([]), "provider"),
            worker_ref=None,
            parent=parent_ref,
            _skip_tunnel=True,
        ),
        "instance",
    )

    ref.tell(Preempted())
    await asyncio.sleep(0.1)

    assert len(parent_msgs) == 1
    assert isinstance(parent_msgs[0], InstanceDied)
    assert parent_msgs[0].instance_id == "i-123"


@pytest.mark.asyncio
async def test_instance_full_lifecycle(system):
    from skyward.actors.instance import instance_actor

    parent_msgs: list = []
    provider_msgs: list = []
    parent_ref = system.spawn(collector_behavior(parent_msgs), "parent")
    provider_ref = system.spawn(collector_behavior(provider_msgs), "provider")

    ref = system.spawn(
        instance_actor(
            instance_id="i-123",
            provider_ref=provider_ref,
            worker_ref=None,
            parent=parent_ref,
            _skip_tunnel=True,
        ),
        "instance",
    )

    ref.tell(Running(ip="10.0.0.1"))
    await asyncio.sleep(0.1)

    ref.tell(Bootstrapping(phase="setup", status="running"))
    await asyncio.sleep(0.1)

    ref.tell(Bootstrapped())
    await asyncio.sleep(0.1)

    assert any(isinstance(m, InstanceBecameReady) for m in parent_msgs)
    ready_msg = next(m for m in parent_msgs if isinstance(m, InstanceBecameReady))
    assert ready_msg.instance_id == "i-123"
    assert ready_msg.ip == "10.0.0.1"


@pytest.mark.asyncio
async def test_instance_preempted_in_ready(system):
    from skyward.actors.instance import instance_actor

    parent_msgs: list = []
    parent_ref = system.spawn(collector_behavior(parent_msgs), "parent")

    ref = system.spawn(
        instance_actor(
            instance_id="i-123",
            provider_ref=system.spawn(collector_behavior([]), "provider"),
            worker_ref=None,
            parent=parent_ref,
            _skip_tunnel=True,
        ),
        "instance",
    )

    ref.tell(Running(ip="10.0.0.1"))
    await asyncio.sleep(0.1)
    ref.tell(Bootstrapped())
    await asyncio.sleep(0.1)
    parent_msgs.clear()

    ref.tell(Preempted())
    await asyncio.sleep(0.1)

    assert len(parent_msgs) == 1
    assert isinstance(parent_msgs[0], InstanceDied)



"""RunPod Provider Actor - Casty behavior for RunPod GPU Pods.

Story: idle -> active -> stopped

idle: accepts ClusterRequested, registers SSH key, resolves GPU,
      optionally creates Instant Cluster, emits ClusterProvisioned.
active: handles InstanceRequested, BootstrapRequested, ShutdownRequested.
stopped: all pods terminated.
"""

from __future__ import annotations

import uuid
from contextlib import suppress
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from casty import ActorContext, ActorRef, Behavior, Behaviors
from loguru import logger

from skyward.actors.messages import (
    BootstrapDone,
    BootstrapRequested,
    ClusterProvisioned,
    ClusterRequested,
    InstanceBootstrapped,
    InstanceLaunched,
    InstanceRequested,
    InstanceRunning,
    ProviderMsg,
    ShutdownCompleted,
    ShutdownRequested,
)
from skyward.providers.ssh_keys import get_local_ssh_key, get_ssh_key_path
from skyward.providers.wait import wait_for_ready

from .client import RunPodClient, RunPodError, get_api_key
from .config import CloudType, RunPod
from .state import RunPodClusterState
from .types import (
    ClusterCreateParams,
    CpuPodCreateParams,
    PodCreateParams,
    PodResponse,
    get_gpu_model,
    get_ssh_port,
)

if TYPE_CHECKING:
    from skyward.api.spec import PoolSpec


@dataclass(frozen=True, slots=True)
class _PodRunning:
    event: InstanceRunning


@dataclass(frozen=True, slots=True)
class _PodRunningFailed:
    instance_id: str
    error: str


@dataclass(frozen=True, slots=True)
class _BootstrapScriptDone:
    instance_id: str


@dataclass(frozen=True, slots=True)
class _BootstrapScriptFailed:
    instance_id: str
    error: str


# =============================================================================
# Module-level helpers
# =============================================================================


def _get_image_name(spec: PoolSpec) -> str:
    if spec.image and hasattr(spec.image, "container_image"):
        return spec.image.container_image  # type: ignore[reportAttributeAccessIssue]
    return "runpod/pytorch:2.1.0-py3.10-cuda11.8.0-devel-ubuntu22.04"


async def _resolve_gpu_type(config: RunPod, spec: PoolSpec) -> str | None:
    if not spec.accelerator_name:
        return None

    api_key = get_api_key(config.api_key)
    async with RunPodClient(api_key) as client:
        gpu_types = await client.get_gpu_types()

    log = logger.bind(provider="runpod")
    is_secure = config.cloud_type == CloudType.SECURE
    available = [
        g for g in gpu_types
        if (is_secure and g.get("secureCloud")) or (not is_secure and g.get("communityCloud"))
    ]

    requested = spec.accelerator_name.upper()
    for gpu in available:
        display_name = gpu.get("displayName", "").upper()
        gpu_id = gpu.get("id", "").upper()
        if requested in display_name or requested in gpu_id:
            log.info(
                "Selected GPU type {gpu_id} ({name})",
                gpu_id=gpu["id"], name=gpu.get("displayName"),
            )
            return gpu["id"]

    available_names = [g.get("displayName", g["id"]) for g in available]
    raise RuntimeError(
        f"No GPU type matches '{spec.accelerator_name}'. "
        f"Available: {', '.join(available_names)}"
    )


async def _create_gpu_pod(
    client: RunPodClient,
    config: RunPod,
    cluster: RunPodClusterState,
    event: InstanceRequested,
) -> PodResponse:
    use_spot = cluster.spec.allocation in ("spot", "spot-if-available")

    params: PodCreateParams = {
        "name": f"skyward-{cluster.cluster_id}-{event.node_id}",
        "imageName": _get_image_name(cluster.spec),
        "gpuTypeIds": [cluster.gpu_type_id or ""],
        "gpuCount": cluster.spec.accelerator_count or 1,
        "cloudType": config.cloud_type.value.upper(),
        "containerDiskInGb": config.container_disk_gb,
        "volumeInGb": config.volume_gb,
        "volumeMountPath": config.volume_mount_path,
        "ports": list(config.ports),
        "interruptible": use_spot,
    }

    if config.data_center_ids != "global":
        params["dataCenterIds"] = list(config.data_center_ids)

    return await client.create_pod(params)


async def _create_cpu_pod(
    client: RunPodClient,
    config: RunPod,
    cluster: RunPodClusterState,
) -> PodResponse:
    vcpus = cluster.spec.vcpus or 2
    memory_gb = cluster.spec.memory_gb or 2
    disk_gb = min(config.container_disk_gb, 20)
    instance_id = f"cpu{config.cpu_clock}-{vcpus}-{memory_gb}"

    params: CpuPodCreateParams = {
        "instanceId": instance_id,
        "cloudType": config.cloud_type.value.upper(),
        "containerDiskInGb": disk_gb,
        "startSsh": True,
        "templateId": "runpod-ubuntu",
        "ports": ",".join(config.ports),
        "deployCost": cluster.spec.max_hourly_cost or 0.50,
    }

    if config.data_center_ids != "global":
        params["dataCenterId"] = config.data_center_ids[0]

    logger.bind(provider="runpod").info("Creating CPU pod with instance {iid}", iid=instance_id)
    return await client.create_cpu_pod(params)


async def _create_instant_cluster(
    config: RunPod,
    state: RunPodClusterState,
    event: ClusterRequested,
) -> None:
    logger.bind(provider="runpod").info(
        "Creating Instant Cluster with {n} nodes",
        n=event.spec.nodes,
    )

    api_key = get_api_key(config.api_key)
    cluster_type = "TRAINING"
    deploy_cost = event.spec.max_hourly_cost or 10.0

    params: ClusterCreateParams = {
        "clusterName": f"skyward-{state.cluster_id}",
        "gpuTypeId": state.gpu_type_id or "",
        "podCount": event.spec.nodes,
        "gpuCountPerPod": event.spec.accelerator_count or 1,
        "type": cluster_type,
        "imageName": _get_image_name(event.spec),
        "startSsh": True,
        "containerDiskInGb": config.container_disk_gb,
        "volumeInGb": config.volume_gb,
        "volumeMountPath": config.volume_mount_path,
        "ports": ",".join(config.ports),
        "deployCost": deploy_cost,
    }

    if config.data_center_ids != "global":
        params["dataCenterId"] = config.data_center_ids[0]

    try:
        async with RunPodClient(api_key) as client:
            cluster = await client.create_cluster(params)
    except RunPodError as e:
        logger.bind(provider="runpod").error("Failed to create Instant Cluster: {err}", err=e)
        raise

    state.runpod_cluster_id = cluster["id"]
    _log = logger.bind(provider="runpod")
    _log.info("Instant Cluster created with id {cid}", cid=state.runpod_cluster_id)

    for pod in cluster["pods"]:
        pod_id = pod["id"]
        node_id = pod.get("clusterIdx", len(state.pod_ids))
        state.pod_ids[node_id] = pod_id

        cluster_ip = pod.get("clusterIp")
        if cluster_ip:
            state.cluster_ips[node_id] = cluster_ip
            _log.debug(
                "Node {nid} -> pod {pid}, cluster IP {cip}",
                nid=node_id, pid=pod_id, cip=cluster_ip,
            )
        else:
            _log.debug("Node {nid} -> pod {pid} (no cluster IP yet)", nid=node_id, pid=pod_id)


async def _wait_for_running(
    config: RunPod,
    cluster: RunPodClusterState,
    event: InstanceLaunched,
) -> InstanceRunning:
    logger.bind(provider="runpod").info(
        "Waiting for pod {pid} to be running",
        pid=event.instance_id,
    )

    api_key = get_api_key(config.api_key)

    async with RunPodClient(api_key) as client:
        pod = await wait_for_ready(
            poll_fn=lambda: client.get_pod(event.instance_id),
            ready_check=lambda p: (
                p is not None
                and p.get("desiredStatus") == "RUNNING"
                and bool(p.get("publicIp"))
            ),
            terminal_check=lambda p: (
                p is not None and p.get("desiredStatus") == "TERMINATED"
            ),
            timeout=config.provision_timeout,
            interval=5.0,
            description=f"RunPod pod {event.instance_id}",
        )

    if not pod:
        raise RuntimeError(f"Pod {event.instance_id} not found after polling")

    ip = pod.get("publicIp") or ""
    ssh_port = get_ssh_port(pod)
    hourly_rate = pod.get("costPerHr", 0.0)
    adjusted_rate = pod.get("adjustedCostPerHr", hourly_rate)

    private_ip: str | None = None
    if cluster.is_instant_cluster:
        private_ip = cluster.cluster_ips.get(event.node_id)
        if private_ip:
            logger.bind(provider="runpod").debug(
                "Using cluster IP {ip} for node {nid}",
                ip=private_ip, nid=event.node_id,
            )

    cluster.hourly_rate = adjusted_rate
    cluster.on_demand_rate = hourly_rate
    cluster.gpu_count = pod.get("gpuCount", 0)
    cluster.gpu_model = cluster.spec.accelerator_name or get_gpu_model(pod)
    cluster.vcpus = pod.get("vcpuCount", 0)
    cluster.memory_gb = pod.get("memoryInGb", 0.0)

    return InstanceRunning(
        request_id=event.request_id,
        cluster_id=event.cluster_id,
        node_id=event.node_id,
        provider="runpod",
        instance_id=event.instance_id,
        ip=ip,
        private_ip=private_ip,
        ssh_port=ssh_port,
        spot=pod.get("interruptible", False),
        hourly_rate=adjusted_rate,
        on_demand_rate=hourly_rate,
        billing_increment=1,
        instance_type=cluster.gpu_type_id or "",
        gpu_count=cluster.gpu_count,
        gpu_model=cluster.gpu_model,
        vcpus=cluster.vcpus,
        memory_gb=cluster.memory_gb,
        gpu_vram_gb=cluster.gpu_vram_gb,
        region=cluster.region,
        ssh_user=cluster.username,
        ssh_key_path=cluster.ssh_key_path,
    )


async def _run_bootstrap(
    config: RunPod,
    cluster: RunPodClusterState,
    event: BootstrapRequested,
) -> str:
    info = event.instance
    ttl = cluster.spec.ttl or config.instance_timeout
    bootstrap_script = cluster.spec.image.generate_bootstrap(
        ttl=ttl,
        shutdown_command=(
            "eval $(cat /proc/1/environ | tr '\\0' '\\n' "
            "| grep RUNPOD_ | sed 's/^/export /'); "
            "runpodctl remove pod $RUNPOD_POD_ID"
        ),
    )

    logger.bind(provider="runpod").info(
        "Connecting to {ip}:{port} to run bootstrap",
        ip=info.ip, port=info.ssh_port,
    )
    from skyward.providers.bootstrap import run_bootstrap_via_ssh, wait_for_ssh

    transport = await wait_for_ssh(
        host=info.ip,
        user=cluster.username,
        key_path=cluster.ssh_key_path,
        port=info.ssh_port,
        timeout=60.0,
    )

    try:
        await transport.run("mkdir -p /opt/skyward")
        await run_bootstrap_via_ssh(
            transport=transport,
            info=info,
            bootstrap_script=bootstrap_script,
            log_prefix="RunPod: ",
        )
    finally:
        await transport.close()

    logger.bind(provider="runpod").debug("Bootstrap script launched on {iid}", iid=info.id)
    return info.id


# =============================================================================
# Actor behavior
# =============================================================================


def runpod_provider_actor(
    config: RunPod,
) -> Behavior[ProviderMsg]:
    log = logger.bind(provider="runpod")

    def idle() -> Behavior[ProviderMsg]:
        async def receive(
            ctx: ActorContext[ProviderMsg], msg: ProviderMsg,
        ) -> Behavior[ProviderMsg]:
            match msg:
                case ClusterRequested(
                    provider="runpod", reply_to=caller,
                ) as event:
                    cluster_id = f"runpod-{uuid.uuid4().hex[:8]}"

                    initial_region = (
                        config.data_center_ids[0]
                        if config.data_center_ids != "global"
                        else "global"
                    )
                    state = RunPodClusterState(
                        cluster_id=cluster_id,
                        spec=event.spec,
                        cloud_type=config.cloud_type.value,
                        data_center_ids=config.data_center_ids,
                        region=initial_region,
                    )

                    ssh_key_path = get_ssh_key_path()
                    _, ssh_public_key = get_local_ssh_key()
                    state.ssh_key_path = ssh_key_path
                    state.ssh_public_key = ssh_public_key

                    api_key = get_api_key(config.api_key)
                    async with RunPodClient(api_key) as client:
                        await client.ensure_ssh_key(ssh_public_key)
                        log.debug("SSH key registered on account")

                    gpu_type_id = await _resolve_gpu_type(config, event.spec)
                    state.gpu_type_id = gpu_type_id

                    if event.spec.nodes >= 2:
                        await _create_instant_cluster(config, state, event)

                    provisioned = ClusterProvisioned(
                        request_id=event.request_id,
                        cluster_id=cluster_id,
                        provider="runpod",
                    )
                    if caller:
                        caller.tell(provisioned)

                    return active(state)

                case _:
                    return Behaviors.same()

        return Behaviors.receive(receive)

    def active(
        state: RunPodClusterState,
        node_refs: dict[int, ActorRef] | None = None,
    ) -> Behavior[ProviderMsg]:
        refs = node_refs or {}

        async def receive(
            ctx: ActorContext[ProviderMsg], msg: ProviderMsg,
        ) -> Behavior[ProviderMsg]:
            match msg:
                case InstanceRequested(
                    provider="runpod", reply_to=node_ref,
                ) as event:
                    if event.cluster_id != state.cluster_id:
                        return Behaviors.same()

                    new_refs = {**refs}
                    if node_ref:
                        new_refs[event.node_id] = node_ref

                    if state.is_instant_cluster:
                        pod_id = state.pod_ids.get(event.node_id)
                        if pod_id:
                            log.info(
                                "Instant Cluster pod {pid} for node {nid}",
                                pid=pod_id, nid=event.node_id,
                            )
                            state.pending_nodes.add(event.node_id)
                            launched = InstanceLaunched(
                                request_id=event.request_id,
                                cluster_id=event.cluster_id,
                                node_id=event.node_id,
                                provider="runpod",
                                instance_id=pod_id,
                            )
                            ctx.pipe_to_self(
                                coro=_wait_for_running(config, state, launched),
                                mapper=lambda result: _PodRunning(event=result),  # type: ignore[reportArgumentType]
                                on_failure=lambda e: _PodRunningFailed(  # type: ignore[reportArgumentType]
                                    instance_id=pod_id,
                                    error=str(e),
                                ),
                            )
                        return active(state, new_refs)

                    log.info("Launching pod for node {nid}", nid=event.node_id)
                    api_key = get_api_key(config.api_key)

                    try:
                        async with RunPodClient(api_key) as client:
                            if state.gpu_type_id:
                                pod = await _create_gpu_pod(client, config, state, event)
                            else:
                                pod = await _create_cpu_pod(client, config, state)
                    except RunPodError as e:
                        log.error("Failed to create pod: {err}", err=e)
                        return Behaviors.same()

                    pod_id = pod["id"]
                    log.debug("Pod created with id {pid}", pid=pod_id)
                    state.pod_ids[event.node_id] = pod_id
                    state.pending_nodes.add(event.node_id)
                    machine = pod.get("machine") or {}
                    state.region = machine.get("dataCenterId") or machine.get("location") or ""

                    launched = InstanceLaunched(
                        request_id=event.request_id,
                        cluster_id=event.cluster_id,
                        node_id=event.node_id,
                        provider="runpod",
                        instance_id=pod_id,
                    )
                    ctx.pipe_to_self(
                        coro=_wait_for_running(config, state, launched),
                        mapper=lambda result: _PodRunning(event=result),  # type: ignore[reportArgumentType]
                        on_failure=lambda e: _PodRunningFailed(  # type: ignore[reportArgumentType]
                            instance_id=pod_id,
                            error=str(e),
                        ),
                    )
                    return active(state, new_refs)

                case BootstrapRequested(instance=info, cluster_id=cid) if info.provider == "runpod":
                    if cid != state.cluster_id:
                        return Behaviors.same()
                    ctx.pipe_to_self(
                        coro=_run_bootstrap(config, state, msg),
                        mapper=lambda instance_id: _BootstrapScriptDone(instance_id=instance_id),  # type: ignore[reportArgumentType]
                        on_failure=lambda e: _BootstrapScriptFailed(  # type: ignore[reportArgumentType]
                            instance_id=info.id,
                            error=str(e),
                        ),
                    )
                    return Behaviors.same()

                case BootstrapDone(instance=info, success=True):
                    if (target := refs.get(info.node)):
                        target.tell(InstanceBootstrapped(instance=info))
                    return Behaviors.same()

                case BootstrapDone(success=False):
                    return Behaviors.same()

                case _PodRunning(event=running_event):
                    target = refs.get(running_event.node_id)
                    if target:
                        target.tell(running_event)
                    return Behaviors.same()

                case _PodRunningFailed(instance_id=iid, error=err):
                    log.error("Pod {pid} did not become ready: {err}", pid=iid, err=err)
                    return Behaviors.same()

                case _BootstrapScriptDone(instance_id=iid):
                    log.debug("Bootstrap script dispatched on {iid}", iid=iid)
                    return Behaviors.same()

                case _BootstrapScriptFailed(instance_id=iid, error=err):
                    log.error("Bootstrap script failed on {iid}: {err}", iid=iid, err=err)
                    return Behaviors.same()

                case ShutdownRequested(
                    cluster_id=cid, reply_to=reply_to,
                ):
                    if cid != state.cluster_id:
                        return Behaviors.same()

                    log.info("Shutting down cluster {cid}", cid=state.cluster_id)
                    api_key = get_api_key(config.api_key)

                    async with RunPodClient(api_key) as client:
                        if state.is_instant_cluster:
                            with suppress(Exception):
                                await client.delete_cluster(state.runpod_cluster_id)  # type: ignore[arg-type]
                        else:
                            for pod_id in state.pod_ids.values():
                                with suppress(Exception):
                                    await client.terminate_pod(pod_id)

                    if reply_to is not None:
                        reply_to.tell(ShutdownCompleted(cluster_id=state.cluster_id))
                    return Behaviors.stopped()

                case _:
                    return Behaviors.same()

        return Behaviors.receive(receive)

    return idle()



"""Infrastructure and cluster panels.

Components for the header/footer of the panel.
"""

from __future__ import annotations

import time

from rich.console import RenderableType
from rich.table import Table
from rich.text import Text

from ..viewmodel import ClusterVM, HeaderVM, InfraVM

SPINNER_FRAMES = ("⠋", "⠙", "⠹", "⠸", "⠼", "⠴", "⠦", "⠧", "⠇", "⠏")


def _spinner_frame() -> str:
    idx = int(time.monotonic() * 10) % len(SPINNER_FRAMES)
    return SPINNER_FRAMES[idx]


def _format_duration(seconds: float) -> str:
    mins, secs = divmod(int(seconds), 60)
    return f"{mins}:{secs:02d}"


class HeaderBar:
    """Single-line header: branding + infra info.

    ● s k y w a r d   aws > us-east-1 > g5g.xlarge · 4 vCPU · 8 GB · 2 spot @ $0.27/hr
    """

    def __init__(self, header: HeaderVM, infra: InfraVM | None) -> None:
        self._header = header
        self._infra = infra

    def render(self, width: int) -> RenderableType:
        vm = self._header
        has_active = any(s == "in_progress" for s in vm.phases.values())
        all_done = all(s == "completed" for s in vm.phases.values())

        match (all_done, has_active):
            case (True, _):
                marker, marker_style = "", "green bold"
            case (_, True):
                marker = "" if vm.blink_on else ""
                marker_style = "cyan bold"
            case _:
                marker, marker_style = "", "dim"

        # Left: branding
        left = Text()
        left.append(f"{marker} ", style=marker_style)
        left.append("s k y w a r d", style="bold")

        infra = self._infra

        # Center: provider > region > instance · specs
        center = Text()
        if infra is not None:
            spinner = _spinner_frame()
            center.append(infra.provider, style="yellow")
            center.append(" > ", style="dim")
            center.append(infra.region)
            center.append(" > ", style="dim")
            itype_style = "bold" if infra.instance_type else "dim"
            center.append(
                infra.instance_type or spinner, style=itype_style,
            )

            if infra.vcpus > 0:
                center.append(" · ", style="dim")
                if infra.gpu_info:
                    center.append(infra.gpu_info, style="magenta")
                    center.append(" · ", style="dim")
                center.append(f"{infra.vcpus} vCPU", style="white")
                center.append(" · ", style="dim")
                center.append(f"{infra.memory_gb} GB", style="white")
            elif infra.gpu_info:
                center.append(" · ", style="dim")
                center.append(infra.gpu_info, style="magenta")

        # Right: allocation @ rate
        right = Text()
        if infra is not None:
            spinner = _spinner_frame()
            right.append(infra.allocation, style="cyan")
            right.append(" @ ", style="dim")
            if infra.hourly_rate == "$0.00/hr":
                right.append(spinner, style="dim")
            else:
                right.append(infra.hourly_rate, style="green bold")

        table = Table(box=None, show_header=False, padding=0, expand=True)
        table.add_column(justify="left")
        table.add_column(justify="center", ratio=1)
        table.add_column(justify="right")
        table.add_row(left, center, right)

        return table


class FooterBar:
    """Single-line footer: cluster metrics (left) + cost/elapsed (right).

    cpu ▂▂ 21%  gpu ▁▁ 0%  mem ▁▁ 10%  vram ▁▁ 0%      estimated total: ~$0.42  elapsed: 5:23
    """

    def __init__(self, cluster: ClusterVM, header: HeaderVM) -> None:
        self._cluster = cluster
        self._header = header

    def render(self) -> Text:
        vm = self._header
        cluster = self._cluster
        text = Text()

        text.append("estimated total: ", style="dim")
        text.append(f"~${vm.cost:.2f}", style="green bold")
        text.append(" · ", style="dim")
        text.append("elapsed: ", style="dim")
        text.append(_format_duration(vm.elapsed_seconds), style="white")
        text.append(" · ", style="dim")
        text.append("averages: ", style="dim")

        for i, metric in enumerate((cluster.cpu, cluster.gpu, cluster.mem, cluster.gpu_mem)):
            if i > 0:
                text.append(", ", style="dim")
            text.append(f"{metric.label} ", style="dim")
            text.append(f"{metric.value:.0f}{metric.unit}", style=f"{metric.style} bold")

        return text


def create_header(header: HeaderVM, infra: InfraVM | None, width: int) -> RenderableType:
    return HeaderBar(header, infra).render(width)


def create_footer(cluster: ClusterVM, header: HeaderVM) -> Text:
    return FooterBar(cluster, header).render()



def test_dict_setitem_getitem(registry):
    proxy = registry.dict("test_dict_basic")

    proxy["key1"] = "value1"
    assert proxy["key1"] == "value1"

    proxy["key2"] = 42
    assert proxy["key2"] == 42


def test_dict_contains(registry):
    proxy = registry.dict("test_dict_contains")

    proxy["exists"] = True
    assert "exists" in proxy
    assert "missing" not in proxy


def test_dict_delete(registry):
    proxy = registry.dict("test_dict_del")

    proxy["key"] = "value"
    del proxy["key"]
    assert "key" not in proxy


def test_dict_get_default(registry):
    proxy = registry.dict("test_dict_get")

    assert proxy.get("missing") is None
    assert proxy.get("missing", "default") == "default"
    proxy["key"] = "value"
    assert proxy.get("key") == "value"


def test_dict_update(registry):
    proxy = registry.dict("test_dict_update")

    proxy.update({"a": 1, "b": 2, "c": 3})
    assert proxy["a"] == 1
    assert proxy["b"] == 2
    assert proxy["c"] == 3


def test_dict_pop(registry):
    proxy = registry.dict("test_dict_pop")

    proxy["key"] = "value"
    result = proxy.pop("key")
    assert result == "value"
    assert "key" not in proxy
    assert proxy.pop("missing", "default") == "default"



"""Metric visualization components: Sparkline and MetricBadge."""

from __future__ import annotations

from rich.text import Text

from ..viewmodel import MetricVM

# Sparkline characters (8 levels from empty to full)
_SPARKLINE_CHARS = "▁▂▃▄▅▆▇█"


class Sparkline:
    """Renders metric history as unicode sparkline.

    Converts a sequence of percentage values (0-100) into a visual
    sparkline using unicode block characters.

    Example output: ▁▂▃▄▅▆▇█
    """

    def __init__(self, history: tuple[float, ...], width: int = 8, style: str = "green") -> None:
        self._history = history
        self._width = width
        self._style = style

    def render(self) -> Text:
        """Render sparkline to Rich Text."""
        if not self._history:
            return Text(_SPARKLINE_CHARS[0] * self._width, style=self._style)

        # Take last `width` values
        values = list(self._history)[-self._width :]

        # Pad with first char if not enough values
        while len(values) < self._width:
            values.insert(0, 0.0)

        chars = [
            _SPARKLINE_CHARS[min(int(max(0.0, min(100.0, v)) / 100.0 * 7), 7)]
            for v in values
        ]

        return Text("".join(chars), style=self._style)


class MetricBadge:
    """Renders a labeled metric with sparkline and current value.

    Example output: cpu ▁▂▃▄▅▆▇█ 78%
    """

    def __init__(self, vm: MetricVM) -> None:
        self._vm = vm

    def render(self) -> Text:
        """Render metric badge to Rich Text."""
        vm = self._vm
        result = Text()

        # Label (expected to be 4 chars, e.g. "cpu ", "mem ", "gpu ", "temp")
        result.append(vm.label, style="dim")
        result.append(" ")  # separator

        # Sparkline
        sparkline = Sparkline(vm.history, width=vm.spark_width, style=vm.style)
        result.append_text(sparkline.render())

        # Current value
        result.append(f" {vm.value:>2.0f}{vm.unit}", style=f"{vm.style} bold")

        return result

    def render_inline(self) -> Text:
        vm = self._vm
        result = Text()
        result.append(vm.label, style="dim")
        result.append(" ")
        sparkline = Sparkline(vm.history, width=vm.spark_width, style=vm.style)
        result.append_text(sparkline.render())
        result.append(f" {vm.value:>2.0f}{vm.unit}", style=f"{vm.style} bold")
        return result


def temp_color(temp: float) -> str:
    """Return color based on temperature: green < 60, yellow 60-75, red > 75."""
    if temp < 60:
        return "green"
    if temp < 75:
        return "yellow"
    return "red"
